import pytest
import socket
import asyncio
import contextlib
from panoramisk import Manager
from panoramisk import utils
from panoramisk.actions import Action


def unused_tcp_port_factory():
    """Find an unused localhost TCP port from 1024-65535 and return it."""
    with contextlib.closing(socket.socket()) as sock:
        sock.bind(('127.0.0.1', 0))
        return sock.getsockname()[1]


def message(msg):
    EOL = utils.EOL.encode('utf8')
    return EOL.join([l.strip() for l in msg.strip().split(b'\n')]) + EOL + EOL


def login(action_id):
    return message(b'''
    Response: Success
    ActionID: ''' + action_id.encode('utf8') + b'''
    Message: Authentication accepted
    ''')


def resp(action_id):
    return message(b'''
    Response: Success
    ActionID: ''' + action_id.encode('utf8') + b'''
    Message: Resp
    ''')


def pong(action_id):
    return message(b'''
    Response: Success
    ActionID: ''' + action_id.encode('utf8') + b'''
    Ping: Pong
    Timestamp: 1409169929.412068
    ''')


class Event:

    def __init__(self, event):
        self.event = event


class _Asterisk(asyncio.Protocol):

    def connection_made(self, transport):
        self.uuid = transport.get_extra_info('peername')
        self.transport = transport
        self.factory.clients[self.uuid] = self
        if not self.factory.started.done():
            self.factory.started.set_result(True)

    def data_received(self, data):
        data = data.decode('utf8')
        data = data.split(utils.EOL + utils.EOL)
        if data:
            if not self.factory._data_received.done():
                self.factory._data_received.set_result(True)
        for line in data:
            if not line.strip():
                continue
            a = Action()
            for kv in line.split('\n'):
                if ':' in kv:
                    k, v = kv.split(':', 1)
                    a[k] = v.strip()
            self.factory.actions.append(a)
            print(repr(a))  # debug
            if 'action' in a:
                if a['action'] == 'Login':
                    message = login(a['actionid'])
                elif a['action'] == 'Ping':
                    message = pong(a['actionid'])
                else:
                    message = resp(a['actionid'])
                print(repr(message))  # debug
                self.transport.write(message)


class Asterisk(object):

    clients = {}

    def __init__(self, loop, port):
        self.loop = loop
        self.port = port
        self.protocol = None

    @property
    def client(self):
        return list(self.clients.values())[0]

    async def start(self):
        self.actions = []
        self.clients = {}
        self.started = asyncio.Future()
        self._data_received = asyncio.Future()
        self.protocol = _Asterisk()
        self.protocol.factory = self
        await self.loop.create_server(
            lambda: self.protocol, '127.0.0.1', self.port)
        return self.started

    async def data_received(self):
        self._data_received = asyncio.Future()
        await self._data_received
        return self._data_received

    async def stop(self):
        for client in self.clients.values():
            client.transport.close()


@pytest.mark.asyncio
async def test_reconnection_without_lost(event_loop):
    unused_tcp_port = unused_tcp_port_factory()

    print('Start server on %s' % unused_tcp_port)
    manager = Manager(loop=event_loop,
                      host='127.0.0.1',
                      port=unused_tcp_port,
                      username='user',
                      secret='secret')
    server = Asterisk(event_loop, unused_tcp_port)

    await server.start()
    await manager.connect()
    await server.data_received()
    print(server.actions)
    login = server.actions[0]
    assert login['action'] == 'Login'
    unused_tcp_port = unused_tcp_port_factory()

    print('Restart server on %s' % unused_tcp_port)
    server.port = unused_tcp_port
    manager.config['port'] = unused_tcp_port
    await server.stop()

    manager.send_action({'Action': 'Ping'})
    f = manager.send_action({'Action': 'Test', 'Command': 'test'})
    fully_booted = Event('FullyBooted')

    await asyncio.sleep(.1)
    assert manager.awaiting_actions
    await server.start()
    await asyncio.sleep(2)
    assert manager.awaiting_actions
    manager.dispatch(fully_booted)
    await asyncio.sleep(.5)
    assert not manager.awaiting_actions
    resp = await f
    await server.stop()

    login2 = server.actions[0]
    assert login2['action'] == 'Login'
    assert login['actionid'] != login2['actionid']

    test_action = server.actions[-1]
    assert test_action.id == resp.id
    assert test_action['action'] == 'Test'



"""
Example script to originate a call through Asterisk Manager Interface.

Usage: python originate.py config.ini

"""
import sys
import asyncio
from panoramisk.call_manager import CallManager


async def originate():
    callmanager = CallManager.from_config(sys.argv[1])
    await callmanager.connect()
    call = await callmanager.send_originate({
        'Action': 'Originate',
        'Channel': 'Local/gpasgrimaud@bearstech',
        'WaitTime': 20,
        'CallerID': 'gawel',
        'Exten': '1000',
        'Context': 'ael-demo',
        'Priority': 1})
    print(call)
    while not call.queue.empty():
        event = call.queue.get_nowait()
        print(event)
    while True:
        event = await call.queue.get()
        print(event)
        if event.event.lower() == 'hangup' and event.cause in ('0', '17'):
            break
    callmanager.clean_originate(call)
    callmanager.close()


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(originate())
    loop.close()


if __name__ == '__main__':
    main()



import asyncio
from panoramisk import Manager

manager = Manager(loop=asyncio.get_event_loop(),
                  host='ip',
                  username='user',
                  secret='secret')


@manager.register_event('*')
def callback(manager, message):
    if "FullyBooted" not in message.event:
        """This will print every event, but the FullyBooted events as these
        will continuously spam your screen"""
        print(message)


"""
# This will print NewChannel Events
@manager.register_event('NewChannel')
def callback(manager, message):
    print(message)


# This will print Hangup Events
@manager.register_event('Hangup')
def callback(manager, message):
    print(message)
"""


def main():
    manager.connect()
    try:
        manager.loop.run_forever()
    except KeyboardInterrupt:
        manager.loop.close()


if __name__ == '__main__':
    main()



import asyncio
import pytest
from panoramisk.fast_agi import Application

FAST_AGI_PAYLOAD = b'''agi_network: yes
agi_network_script: call_waiting
agi_request: agi://127.0.0.1:4574/call_waiting
agi_channel: SIP/xxxxxx-00000000
agi_language: en_US
agi_type: SIP
agi_uniqueid: 1437920906.0
agi_version: asterisk
agi_callerid: 201
agi_calleridname: user 201
agi_callingpres: 0
agi_callingani2: 0
agi_callington: 0
agi_callingtns: 0
agi_dnid: 9011
agi_rdnis: unknown
agi_context: default
agi_extension: 9011
agi_priority: 2
agi_enhanced: 0.0
agi_accountcode: default
agi_threadid: -1260881040
agi_arg_1: answered

'''


async def call_waiting(request):
    r = await request.send_command('ANSWER')
    v = {'msg': '', 'result': ('0', ''),
         'status_code': 200}
    assert r == v


async def fake_asterisk_client(loop, unused_tcp_port):
    reader, writer = await asyncio.open_connection(
        '127.0.0.1', unused_tcp_port)
    # send headers
    writer.write(FAST_AGI_PAYLOAD)
    # read it back
    msg_back = await reader.readline()
    writer.write(b'100 Trying...\n')
    writer.write(b'200 result=0\n')
    writer.close()
    return msg_back


@pytest.mark.asyncio
async def test_fast_agi_application(event_loop, unused_tcp_port):
    fa_app = Application(loop=event_loop)
    fa_app.add_route('call_waiting', call_waiting)

    server = await asyncio.start_server(fa_app.handler, '127.0.0.1',
                                        unused_tcp_port)

    msg_back = await fake_asterisk_client(event_loop,
                                          unused_tcp_port)
    assert b'ANSWER\n' == msg_back

    server.close()
    await server.wait_closed()
    await asyncio.sleep(1)  # Wait the end of endpoint



from panoramisk import utils


def test_parse_agi_valid_result():
    res = utils.parse_agi_result('200 result=0')
    assert res == {'msg': '', 'result': ('0', ''), 'status_code': 200}

    res = utils.parse_agi_result('200 result=1')
    assert res == {'msg': '', 'result': ('1', ''), 'status_code': 200}

    res = utils.parse_agi_result('200 result=1234')
    assert res == {'msg': '', 'result': ('1234', ''), 'status_code': 200}

    res = utils.parse_agi_result('200 result= (timeout)')
    assert res == {'msg': '', 'result': ('', ''), 'status_code': 200}


def test_parse_agi_invalid_result():
    res = utils.parse_agi_result('510 Invalid or unknown command')
    assert res == {'msg': '', 'result': ('', ''),
                   'error': 'AGIInvalidCommand',
                   'status_code': 510}

    res = utils.parse_agi_result('520 Use this')
    assert res == {'msg': '520 Use this', 'result': ('', ''),
                   'error': 'AGIUsageError',
                   'status_code': 520}



from . import utils
from urllib.parse import unquote


class Message(utils.CaseInsensitiveDict):
    """Handle both Responses and Events with the same api:

    ..
        >>> resp = Message({'Response': 'Follows'}, 'Response body')
        >>> event = Message({'Event': 'MeetmeEnd', 'Meetme': '4242'})

    Responses:

    .. code-block:: python

        >>> bool(resp.success)
        True
        >>> resp
        <Message Response='Follows' content='Response body'>
        >>> print(resp.content)
        Response body
        >>> for line in resp.iter_lines():
        ...     print(resp.content)
        Response body

    Events:

    .. code-block:: python

        >>> print(event['meetme'])
        4242
        >>> print(event.meetme)
        4242
        >>> event.unknown_header
        ''

    """

    quoted_keys = ['result']
    success_responses = ['Success', 'Follows', 'Goodbye']

    def __init__(self, headers, content=''):
        super(Message, self).__init__(headers, content=content)
        self.manager = None

    @property
    def id(self):
        if 'commandid' in self:
            return self['commandid']
        elif 'actionid' in self:
            return self['actionid']
        return None

    @property
    def action_id(self):
        if 'actionid' in self:
            return self['actionid']
        return None

    @property
    def success(self):
        """return True if a response status is Success or Follows:

        .. code-block:: python

            >>> resp = Message({'Response': 'Success'})
            >>> print(resp.success)
            True
            >>> resp['Response'] = 'Failed'
            >>> resp.success
            False
        """
        if 'event' in self:
            return True
        if self.response in self.success_responses:
            return True
        return False

    def __repr__(self):
        message = ' '.join(['%s=%r' % i for i in sorted(self.items())])
        return '<Message {0}>'.format(message)

    def iter_lines(self):
        """Iter over response body"""
        for line in self.content.split('\n'):
            yield line

    def parsed_result(self):
        """Get parsed result of AGI command"""
        if 'Result' in self:
            return utils.parse_agi_result(self['Result'])
        else:
            raise ValueError('No result in %r' % self)

    def getdict(self, key):
        """Convert a multi values header to a case-insensitive dict:

        .. code-block:: python

            >>> resp = Message({
            ...     'Response': 'Success',
            ...     'ChanVariable': [
            ...         'FROM_DID=', 'SIPURI=sip:42@10.10.10.1:4242'],
            ... })
            >>> print(resp.chanvariable)
            ['FROM_DID=', 'SIPURI=sip:42@10.10.10.1:4242']
            >>> value = resp.getdict('chanvariable')
            >>> print(value['sipuri'])
            sip:42@10.10.10.1:4242
        """
        values = self.get(key, None)
        if not isinstance(values, list):
            raise TypeError("{0} must be a list. got {1}".format(key, values))
        result = utils.CaseInsensitiveDict()
        for item in values:
            k, v = item.split('=', 1)
            result[k] = v
        return result

    @classmethod
    def from_line(cls, line):
        mlines = line.split(utils.EOL)
        headers = {}
        content = ''
        has_body = ('Response: Follows', 'Response: Fail')
        if mlines[0].startswith(has_body):
            content = mlines.pop()
            while not content and mlines:
                content = mlines.pop()
        for mline in mlines:
            if ': ' in mline:
                k, v = mline.split(': ', 1)
                if k.lower() in cls.quoted_keys:
                    v = unquote(v).strip()
                if k in headers:
                    o = headers.setdefault(k, [])
                    if not isinstance(o, list):
                        o = [o]
                    o.append(v)
                    headers[k] = o
                else:
                    headers[k] = v
        if 'Event' in headers or 'Response' in headers:
            return cls(headers, content)



# -*- coding: utf-8 -*-
import os
import pytest
import asyncio
from panoramisk import testing
from panoramisk import message

test_dir = os.path.join(os.path.dirname(__file__), 'fixtures')


@pytest.fixture
def manager(request, event_loop):
    def manager(stream=None, **config):
        config['loop'] = event_loop
        if stream:
            config['stream'] = os.path.join(test_dir, stream)
        return testing.Manager(**config)
    return manager


def test_connection(manager):
    assert isinstance(manager().connect(), asyncio.Task)


def test_ping(manager):
    manager = manager(stream='ping.yaml')
    future = manager.send_action({'Action': 'Ping'})
    assert 'ping' in future.result()


@pytest.mark.asyncio
async def test_login_ok(manager):
    manager = manager(username='xx', secret='xx', stream='login_ok.yaml')
    authenticated = await manager.authenticated_future
    assert authenticated.success is True
    assert manager.login(manager.authenticated_future) is True


@pytest.mark.asyncio
async def test_login_failed(manager):
    manager = manager(username='xx', secret='xx', stream='login_failed.yaml')
    authenticated = await manager.authenticated_future
    assert authenticated.success is False
    assert manager.login(manager.authenticated_future) is False


def test_logoff(manager):
    manager = manager(stream='logoff.yaml')
    future = manager.send_action({'Action': 'logoff'})
    assert future.result().success is True


def test_queue_status(manager):
    manager = manager(stream='queue_status.yaml')
    future = manager.send_action({'Action': 'QueueStatus',
                                  'Queue': 'xxxxxxxxxxxxxxxx-tous'})
    responses = future.result()
    assert len(responses) == 9


def test_pjsip_show_endpoint(manager):
    manager = manager(stream='pjsip_show_endpoint.yaml')
    future = manager.send_action({'Action': 'PJSIPShowEndpoint',
                                  'Endpoint': 'XXXXX'})
    responses = future.result()
    assert len(responses) == 7


def test_command_core_show_version(manager):
    manager = manager(stream='command_core_show_version.yaml')
    future = manager.send_command('core show version')
    responses = future.result()
    assert len(responses) == 4
    # @todo: in responses['content'],
    # you retrieve only '--END COMMAND--' instead of the result of the command


def test_asyncagi_get_variable(manager):
    manager = manager(stream='asyncagi_get_var.yaml')
    future = manager.send_agi_command(
        'SIP/000000-00000a53', 'GET VARIABLE endpoint')
    response = future.result()[-1]
    assert response.result == '200 result=1 (SIP/000000)'
    pretty_result = response.parsed_result()
    assert pretty_result['status_code'] == 200
    assert pretty_result['result'][0] == '1'
    assert pretty_result['result'][1] == 'SIP/000000'


def test_asyncagi_get_variable_on_dead_channel(manager):
    manager = manager(stream='asyncagi_channel_does_not_exist.yaml')
    future = manager.send_agi_command(
        'SIP/eeeeee-00000014', 'GET VARIABLE DIALSTATUS')
    response = future.result()

    assert response.response == 'Error'
    assert response.message == 'Channel SIP/eeeeee-00000014 does not exist.'


def test_originate_sync(manager):
    manager = manager(stream='originate_sync.yaml')
    future = manager.send_action({'Action': 'Originate', 'Async': 'false'})
    response = future.result()
    assert response.success
    assert response.message == 'Originate successfully queued'


def test_close(manager):
    manager().close()


def test_events(manager):
    future = asyncio.Future()

    manager = manager()

    @manager.register_event('Peer*')
    def callback(manager, event):
        future.set_result(event)

    event = message.Message.from_line('Event: PeerStatus')
    assert event.success is True
    assert event['Event'] == 'PeerStatus'
    assert 'Event' in event
    matches = manager.dispatch(event)
    assert matches == ['Peer*']
    assert event is future.result()

    event = message.Message.from_line('Event: NoPeerStatus')
    matches = manager.dispatch(event)
    assert matches == []


def test_coroutine_events_handler(manager):
    async def callback(manager, event):
        # to create quickly a coroutine generator, don't do that on
        # production code
        return

    manager = manager()
    manager.register_event('Peer*', callback)
    event = message.Message.from_line('Event: PeerStatus')
    assert event.success is True
    assert event['Event'] == 'PeerStatus'
    assert 'Event' in event
    matches = manager.dispatch(event)
    assert matches == ['Peer*']


def test_from_config(event_loop, tmpdir):
    f = tmpdir.mkdir("config").join("config.ini")
    f.write('''
[asterisk]
host = 127.0.0.1
user= username
secret = mysecret
    ''')
    manager = testing.Manager.from_config(str(f), loop=event_loop)
    assert manager.config['secret'] == 'mysecret'

    with open(str(f)) as fd:
        manager = testing.Manager.from_config(fd, loop=event_loop)
    assert manager.config['secret'] == 'mysecret'


def test_pinger(manager):
    manager = manager()
    assert isinstance(manager.pinger, asyncio.TimerHandle)
    manager.close()
    assert manager.pinger is None



import asyncio
from . import manager
from . import actions
from datetime import datetime
from functools import partial


class Call:

    def __init__(self, uniqueid):
        self.uniqueid = uniqueid
        self.action_id = None
        self.queue = asyncio.Queue()
        self.created_at = datetime.now()

    def append(self, *events):
        for e in events:
            self.queue.put_nowait(e)


class CallManager(manager.Manager):

    def __init__(self, **config):
        super(CallManager, self).__init__(**config)
        self.CallClass = config.get('CallClass', Call)
        self.calls_queues = {}
        self.calls = {}
        self.register_event('*', self.handle_calls)

    def set_result(self, future, result):
        res = result.result()
        if isinstance(res, (list, tuple)):
            event = res[-1]
        else:
            event = res
        uniqueid = event.uniqueid.split('.', 1)[0]
        if uniqueid in self.calls_queues:
            call = self.calls_queues[uniqueid]
        else:
            call = Call(uniqueid)
        call.action_id = event.action_id
        if not future.done():
            future.set_result(call)

    def send_originate(self, action):
        action['Async'] = 'true'
        action = actions.Action(action)
        future = asyncio.Future()
        self.send_action(action).add_done_callback(
            partial(self.set_result, future))
        return future

    def clean_originate(self, call):
        self.calls_queues.pop(call.uniqueid, None)

    def handle_calls(self, manager, event):
        uniqueid = event.uniqueid or event.uniqueid1
        if uniqueid:
            uniqueid = uniqueid.split('.', 1)[0]
            call = self.calls_queues.setdefault(uniqueid, Call(uniqueid))
            call.append(event)



import asyncio
from panoramisk import Manager
from pprint import pprint


async def queue_status():
    manager = Manager(loop=asyncio.get_event_loop(),
                      host='127.0.0.1', port=5039,
                      username='username', secret='mysecret')
    await manager.connect()
    queues_details = await manager.send_action(
        {'Action': 'QueueStatus', 'Queue': 'queue_name'})
    manager.close()
    pprint(queues_details)


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(queue_status())
    loop.close()


if __name__ == '__main__':
    main()



import asyncio
import logging
from collections import defaultdict
from collections import deque
import re
import fnmatch
from .ami_protocol import AMIProtocol
from . import actions
from . import utils


class Manager:
    """Main object:

    .. code-block:: python

        >>> manager = Manager(
        ...    host='127.0.0.1',
        ...    port=5038,
        ...    ssl=False,
        ...    encoding='utf8')
    """

    defaults = dict(
        host='127.0.0.1',
        port=5038,
        events='on',
        ssl=False,
        encoding='utf8',
        ping_delay=10,
        ping_interval=10,
        reconnect_timeout=2,
        protocol_factory=AMIProtocol,
        save_stream=None,
        loop=None,
        forgetable_actions=('ping', 'login'),
    )

    def __init__(self, **config):
        self.config = dict(self.defaults, **config)
        self.loop = self.config['loop']
        self.log = config.get('log', logging.getLogger(__name__))
        self.callbacks = defaultdict(list)
        self.protocol = None
        self.patterns = []
        self.save_stream = self.config.get('save_stream')
        self.authenticated = False
        self.authenticated_future = None
        self.awaiting_actions = deque()
        self.forgetable_actions = self.config['forgetable_actions']
        self.pinger = None
        self.ping_delay = int(self.config['ping_delay'])
        self.ping_interval = int(self.config['ping_interval'])
        self.reconnect_timeout = int(self.config['reconnect_timeout'])
        self._connected = False
        self.register_event('FullyBooted', self.send_awaiting_actions)
        self.on_login = config.get('on_login', on_login)
        self.on_connect = config.get('on_connect', on_connect)
        self.on_disconnect = config.get('on_disconnect', on_disconnect)

    def connection_made(self, f):
        if getattr(self, 'protocol', None):
            self.protocol.close()
        try:
            transport, protocol = f.result()
        except OSError:  # pragma: no cover
            if self._connected:
                self.log.exception('Not able to connect')
                self._connected = False
            else:
                self.log.warning('Not able to reconnect')
            self.loop.call_later(self.reconnect_timeout, self.connect)
        else:
            self._connected = True
            self.log.debug('Manager connected')
            self.loop.call_soon(self.on_connect, self)
            self.protocol = protocol
            self.protocol.queue = deque()
            self.protocol.factory = self
            self.protocol.log = self.log
            self.protocol.config = self.config
            self.protocol.encoding = self.encoding = self.config['encoding']
            self.responses = self.protocol.responses = {}
            if 'username' in self.config:
                self.authenticated = False
                self.authenticated_future = self.send_action({
                    'Action': 'Login',
                    'Username': self.config['username'],
                    'Secret': self.config['secret'],
                    'Events': self.config['events']})
                self.authenticated_future.add_done_callback(self.login)
            else:
                self.log.debug('username not in config file')
            self.pinger = self.loop.call_later(self.ping_delay, self.ping)

    def login(self, future):
        self.authenticated_future = None
        resp = future.result()
        self.authenticated = bool(resp.success)
        if self.authenticated:
            self.loop.call_soon(self.on_login, self)
        if self.pinger is not None:
            self.pinger.cancel()
        self.pinger = self.loop.call_later(self.ping_delay, self.ping)
        return self.authenticated

    def ping(self):  # pragma: no cover
        self.pinger = self.loop.call_later(self.ping_interval, self.ping)
        self.protocol.send({'Action': 'Ping'})

    async def send_awaiting_actions(self, *_):
        self.log.info('Sending awaiting actions')
        while self.awaiting_actions:
            action = self.awaiting_actions.popleft()
            if action['action'].lower() not in self.forgetable_actions:
                if not action.future.done():
                    self.send_action(action, as_list=action.as_list)

    def send_action(self, action, as_list=None, **kwargs):
        """Send an :class:`~panoramisk.actions.Action` to the server:

        :param action: an Action or dict with action name and parameters to
                       send
        :type action: Action or dict or Command
        :param as_list: If True, the action Future will retrieve all responses
        :type as_list: boolean
        :return: a Future that will receive the response
        :rtype: asyncio.Future

        :Example:

        To retrieve answer in a coroutine::

            manager = Manager()
            resp = await manager.send_action({'Action': 'Status'})

        With a callback::

            manager = Manager()
            future = manager.send_action({'Action': 'Status'})
            future.add_done_callback(handle_status_response)

        See https://wiki.asterisk.org/wiki/display/AST/AMI+Actions for
        more information on actions
        """
        action.update(kwargs)
        return self.protocol.send(action, as_list=as_list)

    def send_command(self, command, as_list=False):
        """Send a :class:`~panoramisk.actions.Command` to the server::

            manager = Manager()
            resp = manager.send_command('http show status')

        Return a response :class:`~panoramisk.message.Message`.
        See https://wiki.asterisk.org/wiki/display/AST/ManagerAction_Command
        """
        action = actions.Action({'Command': command, 'Action': 'Command'},
                                as_list=as_list)
        return self.send_action(action)

    def send_agi_command(self, channel, command, as_list=False):
        """Send a :class:`~panoramisk.actions.Command` to the server:

        :param channel: Channel name where to launch command.
               Ex: 'SIP/000000-00000a53'
        :type channel: String
        :param command: command to launch. Ex: 'GET VARIABLE async_agi_server'
        :type command: String
        :param as_list: If True, the action Future will retrieve all responses
        :type as_list: boolean
        :return: a Future that will receive the response
        :rtype: asyncio.Future

        :Example:

        ::

            manager = Manager()
            resp = manager.send_agi_command('SIP/000000-00000a53',
                                            'GET VARIABLE async_agi_server')


        Return a response :class:`~panoramisk.message.Message`.
        See https://wiki.asterisk.org/wiki/display/AST/Asterisk+11+ManagerAction_AGI
        """
        action = actions.Command({'Action': 'AGI',
                                  'Channel': channel,
                                  'Command': command},
                                 as_list=as_list)
        return self.send_action(action)

    def connect(self, run_forever=False, on_startup=None, on_shutdown=None):
        """connect to the server"""
        if self.loop is None:  # pragma: no cover
            self.loop = asyncio.get_event_loop()
        t = asyncio.Task(
            self.loop.create_connection(
                self.config['protocol_factory'],
                self.config['host'], self.config['port'],
                ssl=self.config['ssl']),
            loop=self.loop)
        t.add_done_callback(self.connection_made)

        if run_forever:
            self.run_forever(on_startup, on_shutdown)
        return t

    def run_forever(self, on_startup, on_shutdown):
        """Start loop forever"""
        try:
            if on_startup:
                self.loop.run_until_complete(on_startup(self))
            self.loop.run_forever()
        except (KeyboardInterrupt, SystemExit):
            self.close()
        finally:
            if on_shutdown:
                self.loop.run_until_complete(on_shutdown(self))
            self.loop.stop()

    def register_event(self, pattern, callback=None):
        """register an event. See :class:`~panoramisk.message.Message`:

        .. code-block:: python

            >>> def callback(manager, event):
            ...     print(manager, event)
            >>> manager = Manager()
            >>> manager.register_event('Meetme*', callback)
            <function callback at 0x...>

        You can also use the manager as a decorator:

        .. code-block:: python

            >>> manager = Manager()
            >>> @manager.register_event('Meetme*')
            ... def callback(manager, event):
            ...     print(manager, event)
        """
        def _register_event(callback):
            if not self.callbacks[pattern]:
                self.patterns.append((pattern,
                                     re.compile(fnmatch.translate(pattern))))
            self.callbacks[pattern].append(callback)
            return callback
        if callback is not None:
            return _register_event(callback)
        else:
            return _register_event

    def dispatch(self, event):
        matches = []
        event.manager = self
        for pattern, regexp in self.patterns:
            match = regexp.match(event.event)
            if match is not None:
                matches.append(pattern)
                for callback in self.callbacks[pattern]:
                    ret = callback(self, event)
                    if (asyncio.iscoroutine(ret) or
                            isinstance(ret, asyncio.Future)):
                        asyncio.ensure_future(ret, loop=self.loop)
        return matches

    def close(self):
        """Close the connection"""
        if self.pinger:
            self.pinger.cancel()
            self.pinger = None
        if getattr(self, 'protocol', None):
            self.protocol.close()

    def connection_lost(self, exc):
        self._connected = False
        self.log.error('Connection lost')
        self.loop.call_soon(self.on_disconnect, self, exc)
        if self.pinger:
            self.pinger.cancel()
            self.pinger = None
        self.log.info('Try to connect again in %d second(s)' % self.reconnect_timeout)
        self.loop.call_later(self.reconnect_timeout, self.connect)

    @classmethod
    def from_config(cls, filename_or_fd, section='asterisk', **kwargs):
        config = utils.config(filename_or_fd, section=section)
        config.update(kwargs)
        return cls(**config)


# noinspection PyUnusedLocal
def on_connect(manager: Manager):
    """
    Callback after connect
    """
    pass


# noinspection PyUnusedLocal
def on_login(manager: Manager):
    """
    Callback after login
    """
    pass


# noinspection PyUnusedLocal
def on_disconnect(manager: Manager, exc: Exception):
    """
    Callback after disconnect
    """
    pass



import logging
import asyncio
from collections import OrderedDict
from .utils import parse_agi_result

log = logging.getLogger(__name__)


class Request:
    def __init__(self, app, headers, reader, writer, encoding='utf-8'):
        self.app = app
        self.headers = headers
        self.reader = reader
        self.writer = writer
        self.encoding = encoding

    async def send_command(self, command):
        """Send a command for FastAGI request:

        :param command: Command to launch on FastAGI request. Ex: 'EXEC StartMusicOnHolds'
        :type command: String

        :Example:

        ::

            async def call_waiting(request):
                print(['AGI variables:', request.headers])
                await request.send_command('ANSWER')
                await request.send_command('EXEC StartMusicOnHold')
                await request.send_command('EXEC Wait 10')

        """
        command += '\n'
        self.writer.write(command.encode(self.encoding))
        await self.writer.drain()

        agi_result = await self._read_result()
        # If Asterisk returns `100 Trying...`, wait for next the response.
        while agi_result.get('status_code') == 100:
            agi_result = await self._read_result()

        # when we got AGIUsageError the following line contains some indication
        if 'error' in agi_result and agi_result['error'] == 'AGIUsageError':
            buff_usage_error = await self.reader.readline()
            agi_result['msg'] += buff_usage_error.decode(self.encoding)

        return agi_result

    async def _read_result(self):
        """Parse read a response from the AGI and parse it.

        :return dict: The AGI response parsed into a dict.
        """
        response = await self.reader.readline()
        return parse_agi_result(response.decode(self.encoding)[:-1])


class Application(dict):
    """Main object:

    .. code-block:: python

        >>> fa_app = Application()
    """

    buf_size = 100

    def __init__(self, default_encoding='utf-8', loop=None):
        super(Application, self).__init__()
        self.default_encoding = default_encoding
        if loop is None:
            loop = asyncio.get_event_loop()
        self.loop = loop
        self._route = OrderedDict()

    def add_route(self, path, endpoint):
        """Add a route for FastAGI requests:

        :param path: URI to answer. Ex: 'calls/start'
        :type path: String
        :param endpoint: command to launch. Ex: start
        :type endpoint: callable

        :Example:

        ::

            async def start(request):
                print('Receive a FastAGI request')
                print(['AGI variables:', request.headers])

            fa_app = Application()
            fa_app.add_route('calls/start', start)

        """
        assert callable(endpoint), endpoint
        if path in self._route:
            raise ValueError('A route already exists.')
        if not asyncio.iscoroutinefunction(endpoint):
            endpoint = asyncio.coroutine(endpoint)
        self._route[path] = endpoint

    def del_route(self, path):
        """Delete a route for FastAGI requests:

        :param path: URI to answer. Ex: 'calls/start'
        :type path: String

        :Example:

        ::

            async def start(request):
                print('Receive a FastAGI request')
                print(['AGI variables:', request.headers])

            fa_app = Application()
            fa_app.add_route('calls/start', start)
            fa_app.del_route('calls/start')

        """
        if path not in self._route:
            raise ValueError('This route doesn\'t exist.')
        del(self._route[path])

    async def handler(self, reader, writer):
        """AsyncIO coroutine handler to launch socket listening.

        :Example:

        ::

            async def start(request):
                print('Receive a FastAGI request')
                print(['AGI variables:', request.headers])

            fa_app = Application()
            fa_app.add_route('calls/start', start)
            coro = asyncio.start_server(fa_app.handler, '0.0.0.0', 4574)
            server = loop.run_until_complete(coro)

        See https://docs.python.org/3/library/asyncio-stream.html
        """
        buffer = b''
        while b'\n\n' not in buffer:
            buffer += await reader.read(self.buf_size)
        lines = buffer[:-2].decode(self.default_encoding).split('\n')
        headers = OrderedDict([
            line.split(': ', 1) for line in lines if ': ' in line
        ])

        agi_network_script = headers.get('agi_network_script')
        log.info('Received FastAGI request from %r for "%s" route',
                 writer.get_extra_info('peername'), agi_network_script)
        log.debug("Asterisk Headers: %r", headers)

        if agi_network_script is not None:
            route = self._route.get(agi_network_script)
            if route is not None:
                request = Request(app=self,
                                  headers=headers,
                                  reader=reader, writer=writer,
                                  encoding=self.default_encoding)
                try:
                    await route(request)
                except BaseException:
                    log.exception(
                        'An exception has been raised for the request "%s"',
                        agi_network_script
                    )
            else:
                log.error('No route for the request "%s"', agi_network_script)
        else:
            log.error('No agi_network_script header for the request')
        log.debug("Closing client socket")
        writer.close()



import logging
import asyncio
from collections import deque

from .message import Message
from . import actions
from . import utils


class AMIProtocol(asyncio.Protocol):

    def connection_made(self, transport):
        self.transport = transport
        self.closed = False
        self.queue = deque()
        self.responses = {}
        self.factory = None
        self.version = None
        self.log = logging.getLogger(__name__)

    def send(self, data, as_list=False):
        encoding = getattr(self, 'encoding', 'ascii')
        if not isinstance(data, actions.Action):
            if 'Command' in data:
                klass = actions.Command
            else:
                klass = actions.Action
            data = klass(data, as_list=as_list)
        data.as_list = as_list
        self.responses[data.id] = data
        if data.action_id:
            self.responses[data.action_id] = data
        try:
            self.transport.write(str(data).encode(encoding))
        except Exception:  # pragma: no cover
            self.log.exception('Fail to send %r' % data)
        return data.future

    def data_received(self, data):
        encoding = getattr(self, 'encoding', 'ascii')
        data = data.decode(encoding, 'ignore')
        if getattr(self.factory, 'save_stream', None):  # pragma: no cover
            stream = self.factory.save_stream
            if hasattr(stream, 'write'):
                stream.write(data.encode(encoding))
            else:
                with open(stream, 'a+') as fd:
                    fd.write(data.encode(encoding))
        # Very verbose, uncomment only if necessary
        # self.log.debug('data received: "%s"', data)

        if self.version is None:
            if data.startswith('Asterisk Call Manager/'):
                version, __, __ = data.partition(utils.EOL)
                __, __, version = version.partition('/')
                self.version = version.strip()
                self.log.info("protocol version: '%s'", self.version)

        if self.queue:
            data = self.queue.popleft() + data
        lines = data.split(utils.EOL+utils.EOL)
        self.queue.append(lines.pop(-1))
        for line in lines:
            # Because sometimes me receive only one EOL from Asterisk
            line = line.strip()
            # Very verbose, uncomment only if necessary
            # self.log.debug('message received: "%s"', line)
            message = Message.from_line(line)
            self.log.debug('message interpreted: %r', message)
            if message is None:
                continue
            self.handle_message(message)

    def handle_message(self, message):
        response = self.responses.get(message.id)
        if response is None and message.action_id:
            response = self.responses.get(message.action_id)
        if response is not None:
            if response.add_message(message):
                # completed; dequeue
                self.responses.pop(response.id)
                if response.action_id:
                    self.responses.pop(response.action_id, None)
        elif 'event' in message:
            if message['event'].lower() == 'shutdown':
                self.connection_lost(message)
            self.factory.dispatch(message)

    def connection_lost(self, exc):
        if not self.closed:
            self.close()
            self.factory.connection_lost(exc)

    def close(self):
        if self.factory and self.responses:
            uuids = set()
            forgetable_actions = self.factory.forgetable_actions
            awaiting_actions = self.factory.awaiting_actions
            for k in list(self.responses.keys()):
                action = self.responses.pop(k)
                uuids.add(action.id)
                if action['action'].lower() in forgetable_actions:
                    continue
                elif action.future.done():  # pragma: no cover
                    continue
                elif action.responses:
                    # If at least one response was receive from asterisk we don't queue it again
                    continue
                else:
                    self.log.info('Adding action "%s" to awaiting list: %s', action['action'].lower(), str(action))
                    awaiting_actions.append(action)
        if not self.closed:
            try:
                self.transport.close()
            finally:
                self.closed = True



import sys
import asyncio
import argparse
import logging
from . import utils
from .call_manager import CallManager
from .call_manager import Call

try:
    import yaml
except ImportError:
    yaml = None


def main(argv=None):
    parser = argparse.ArgumentParser()
    if yaml is None:
        parser.error('You must install pyaml')
    parser.add_argument('-c', '--config',
                        type=argparse.FileType('r'),
                        required=True,
                        help='Config ini file')
    parser.add_argument('-i', '--input',
                        type=argparse.FileType('r'),
                        help='Input yaml file')
    parser.add_argument('-o', '--output',
                        type=argparse.FileType('w'),
                        default='-',
                        help='Stream output file (Default to STDOUT)')
    args = parser.parse_args(argv or sys.argv[1:])
    config = utils.config(args.config)
    config['save_stream'] = args.output
    call_manager = CallManager(**config)
    task = call_manager.connect()

    logging.basicConfig(stream=sys.stderr, level=logging.INFO)
    logging.getLogger('asyncio').setLevel(logging.ERROR)
    log = logging.getLogger('panoramisk')

    def done(future):
        result = future.result()
        log.info('Got result: %r\n', result)
        if isinstance(result, Call):
            while not result.queue.empty():
                print(result.queue.get_nowait())

            def show(f=None):
                if f:
                    print(f.result())
                f = asyncio.Task(result.queue.get())
                f.add_done_callback(show)
            show()

    def send_action(f):
        if args.input:
            action = yaml.load(args.input)
            if action.get('Action').lower() == 'originate':
                future = call_manager.send_originate(action)
            elif 'commandid' in [k.lower() for k in action.keys()]:
                future = call_manager.send_command(action)
            else:
                future = call_manager.send_action(action)
            log.info('Action %r sent', action)
            future.add_done_callback(done)

    def connected(f):
        if call_manager.authenticated_future is not None:
            call_manager.authenticated_future.add_done_callback(send_action)
        else:
            send_action(f)

    task.add_done_callback(connected)

    try:
        call_manager.loop.run_forever()
    except KeyboardInterrupt:
        args.output.close()



import re
import uuid

try:
    from collections.abc import MutableMapping
except ImportError:
    from collections import MutableMapping


from configparser import ConfigParser

EOL = '\r\n'

re_code = re.compile(r'(?P<code>^\d*)\s*(?P<response>.*)')
re_kv = re.compile(r'(?P<key>\w+)=(?P<value>[^\s]+)\s*(?:\((?P<data>.*)\))*')


def parse_agi_result(line):
    """Parse AGI results using Regular expression.

    AGI Result examples::

        100 result=0 Trying...

        200 result=0

        200 result=-1

        200 result=132456

        200 result= (timeout)

        510 Invalid or unknown command

        520-Invalid command syntax. Proper usage follows:
        int() argument must be a string, a bytes-like object or a number, not
        'NoneType'

        HANGUP

    """
    # print("--------------\n", line)
    if line == 'HANGUP':
        return {'error': 'AGIResultHangup',
                'msg': 'User hungup during execution'}

    kwargs = dict(code=0, response="", line=line)
    m = re_code.search(line)
    try:
        kwargs.update(m.groupdict())
    except AttributeError:
        # None has no attribute groupdict
        pass
    return agi_code_check(**kwargs)


def agi_code_check(code=None, response=None, line=None):
    """
    Check the AGI code and return a dict to help on error handling.
    """
    code = int(code)
    response = response or ""
    result = {'status_code': code, 'result': ('', ''), 'msg': ''}
    if code == 100:
        result['msg'] = line
    elif code == 200:
        for key, value, data in re_kv.findall(response):
            result[key] = (value, data)
            # If user hangs up... we get 'hangup' in the data
            if data == 'hangup':
                return {
                    'error': 'AGIResultHangup',
                    'msg': 'User hungup during execution'}
            elif key == 'result' and value == '-1':
                return {
                    'error': 'AGIAppError',
                    'msg': 'Error executing application, or hangup'}
    elif code == 510:
        result['error'] = 'AGIInvalidCommand'
    elif code == 520:
        # AGI Usage error
        result['error'] = 'AGIUsageError'
        result['msg'] = line
    else:
        # Unhandled code or undefined response
        result['error'] = 'AGIUnknownError'
        result['msg'] = line
    return result


class IdGenerator:
    """Generate some uuid for actions:

    .. code-block:: python

        >>> g = IdGenerator('mycounter')

    ..
        >>> IdGenerator.reset(uid='an_uuid4')

    It increments the counter at each calls:

    .. code-block:: python

        >>> print(g())
        mycounter/an_uuid4/1/1
        >>> print(g())
        mycounter/an_uuid4/1/2
    """

    instances = []

    def __init__(self, prefix):
        self.instances.append(self)
        self.prefix = prefix
        self.uid = str(uuid.uuid4())
        self.generator = self.get_generator()

    def get_generator(self):
        i = 0
        max_val = 10000
        while True:
            yield "%s/%s/%d/%d" % (self.prefix,
                                   self.uid, (i // max_val) + 1,
                                   (i % max_val) + 1)
            i += 1

    @classmethod
    def reset(cls, uid=None):
        """Mostly used for unit testing. Allow to use a static uuid and reset
        all counter"""
        for instance in cls.instances:
            if uid:
                instance.uid = uid
            instance.generator = instance.get_generator()

    def get_instances(self):
        """Mostly used for debugging"""
        return ["<%s prefix:%s (uid:%s)>" % (self.__class__.__name__,
                                             i.prefix, self.uid)
                for i in self.instances]

    def __call__(self):
        return next(self.generator)

    def __repr__(self):
        return "<%s prefix:%s (uid:%s)>" % (self.__class__.__name__, self.prefix, self.uid)


class CaseInsensitiveDict(MutableMapping):
    """
    A case-insensitive ``dict``-like object.

    Implements all methods and operations of ``collections.MutableMapping``.

    All keys are expected to be strings. The structure remembers the
    case of the last key to be set, and ``iter(instance)``,
    ``keys()``, ``items()``, ``iterkeys()``, and ``iteritems()``
    will contain case-sensitive keys. However, querying and contains
    testing is case insensitive:

    .. code-block:: python

        cid = CaseInsensitiveDict()
        cid['Action'] = 'SIPnotify'
        cid['aCTION'] == 'SIPnotify'  # True
        list(cid) == ['Action']  # True

    For example, ``event['actionid']`` will return the
    value of a ``'ActionID'`` response event, regardless
    of how the event name was originally stored.
    """
    def __init__(self, data=None, **kwargs):
        self._store = dict()
        self.update(data or {}, **kwargs)

    def __setitem__(self, key, value):
        # Use the lowercased key for lookups, but store the actual
        # key alongside the value.
        self._store[key.lower()] = (key, value)

    def __contains__(self, key):
        return key.lower() in self._store

    def __getattr__(self, attr):
        return self.get(attr, '')

    def __getitem__(self, key):
        return self._store[key.lower()][1]

    def __delitem__(self, key):
        raise NotImplementedError()

    def __iter__(self):
        return (key for key, value in self._store.values())

    def __len__(self):
        return len(self._store)

    def __repr__(self):
        return str(dict(self.items()))


def config(filename_or_fd, section='asterisk'):
    config = ConfigParser()
    if hasattr(filename_or_fd, 'read'):
        if hasattr(config, 'read_file'):
            config.read_file(filename_or_fd)
        else:
            config.readfp(filename_or_fd)
    else:
        config.read(filename_or_fd)
    return dict(config.items(section))



from panoramisk.message import Message
from panoramisk import utils
import pytest


@pytest.fixture
def message():
    def _message(data):
        return Message.from_line(data)
    EOL = utils.EOL
    utils.EOL = '\n'
    yield _message
    utils.EOL = EOL


def test_multivalue(message):
    m = message('''
Event: X
Value: X
Value: Y
''')
    assert m.value == ['X', 'Y']


def test_content(message):
    m = message('''\
Response: Follows
--- blah ---
''')
    assert m.content == '--- blah ---'



from pprint import pprint
import asyncio
from panoramisk import Manager


async def extension_status():
    manager = Manager(loop=asyncio.get_event_loop(),
                      host='127.0.0.1', port=5039,
                      username='user', secret='password')
    await manager.connect()
    extension = await manager.send_action({'Action': 'ExtensionState',
                                           'Exten': '2001',
                                           'Context': 'default'})
    manager.close()
    pprint(extension)


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(extension_status())
    loop.close()


if __name__ == '__main__':
    main()



import asyncio

from . import utils


class Action(utils.CaseInsensitiveDict):
    """Dict like object to handle actions.
    Generate action IDs for you:

    ..
        >>> utils.IdGenerator.reset('myuuid')

    .. code-block:: python

        >>> action = Action({'Action': 'Status'})
        >>> print(action) # doctest: +NORMALIZE_WHITESPACE
        Action: Status
        ActionID: action/myuuid/1/1

        >>> action = Action({'Action': 'SIPnotify',
        ...                  'Variable': ['1', '2']})
        >>> print(action) # doctest: +NORMALIZE_WHITESPACE
        Action: SIPnotify
        ActionID: action/myuuid/1/2
        Variable: 1
        Variable: 2
    """

    action_id_generator = utils.IdGenerator('action')

    def __init__(self, *args, **kwargs):
        self.as_list = kwargs.pop('as_list', None)
        super(Action, self).__init__(*args, **kwargs)
        if 'actionid' not in self:
            self['ActionID'] = self.action_id_generator()
        self.responses = []
        self.future = asyncio.Future()

    @property
    def id(self):
        return self.actionid

    action_id = id

    def __str__(self):
        action = []
        for k, v in sorted(self.items()):
            if isinstance(v, (list, tuple)):
                action.extend(['%s: %s' % (k, i) for i in v])
            else:
                action.append('%s: %s' % (k, v))
        action.append(utils.EOL)
        return utils.EOL.join(action)

    @property
    def multi(self):
        resp = self.responses[0]
        msg = resp.message.lower()
        if self.as_list is not None:
            return bool(self.as_list)
        elif resp.subevent == 'Start':
            return True
        elif 'EventList' in resp and resp['EventList'] == 'start':
            return True
        elif 'will follow' in msg:
            return True
        elif msg.startswith('added') and msg.endswith('to queue'):
            return True
        elif msg.endswith('successfully queued') and self['async'] != 'false':
            return True
        return False

    @property
    def completed(self):
        resp = self.responses[-1]
        if resp.event.endswith('Complete'):
            return True
        elif resp.subevent in ('End', 'Exec'):
            return True
        elif resp.response in ('Success', 'Error', 'Fail', 'Failure'):
            return True
        elif not self.multi:
            return True
        return False

    def add_message(self, message):
        self.responses.append(message)
        multi = self.multi
        if self.completed and not self.future.done():
            if multi and len(self.responses) > 1:
                self.future.set_result(self.responses)
            elif not multi:
                self.future.set_result(self.responses[0])
            else:
                return False
            return True


class Command(Action):
    """Dict like object to handle Commands.
    Generate action/command IDs for you:

    ..
        >>> utils.IdGenerator.reset('myuuid')

    .. code-block:: python

        >>> command = Command({'Command' : 'Do something'})
        >>> print(command) # doctest: +NORMALIZE_WHITESPACE
        Action: Command
        ActionID: action/myuuid/1/1
        Command: Do something
        CommandID: command/myuuid/1/1
    """

    command_id_generator = utils.IdGenerator('command')

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        if 'action' not in self:
            self['Action'] = 'Command'
        if 'commandid' not in self:
            self['CommandID'] = self.command_id_generator()

    @property
    def id(self):
        return self.commandid

    @property
    def action_id(self):
        return self.actionid or None



from pprint import pprint
import asyncio
from panoramisk import fast_agi

loop = asyncio.get_event_loop()


async def call_waiting(request):
    pprint(['AGI variables:', request.headers])

    pprint((await request.send_command('ANSWER')))
    pprint((await request.send_command('SAY DIGITS 1 \"\"')))

    # To Raise a 510 error - 510 Invalid or unknown command
    pprint((await request.send_command('INVALID-COMMAND')))

    # To Raise a 520 error - 520-Invalid command syntax. Proper usage follows:
    pprint((await request.send_command('SAY PHONETIC Hello world .')))

    pprint((await request.send_command('SAY NUMBER 100 \"\"')))
    pprint((await request.send_command('GET DATA hello-world 5000 2')))

    pprint((await request.send_command('EXEC StartMusicOnHold')))
    pprint((await request.send_command('EXEC Wait 30')))


def main():
    fa_app = fast_agi.Application(loop=loop)
    fa_app.add_route('call_waiting', call_waiting)
    coro = asyncio.start_server(fa_app.handler, '0.0.0.0', 4574, loop=loop)
    server = loop.run_until_complete(coro)

    # Serve requests until CTRL+c is pressed
    print('Serving on {}'.format(server.sockets[0].getsockname()))
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    # Close the server
    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close()


if __name__ == '__main__':
    main()



from pprint import pprint
import asyncio
from panoramisk import fast_agi

loop = asyncio.get_event_loop()


async def call_waiting(request):
    pprint(['AGI variables:', request.headers])
    pprint((await request.send_command('ANSWER')))
    pprint((await request.send_command('EXEC StartMusicOnHold')))
    pprint((await request.send_command('EXEC Wait 30')))


def main():
    fa_app = fast_agi.Application(loop=loop)
    fa_app.add_route('call_waiting', call_waiting)
    coro = asyncio.start_server(fa_app.handler, '0.0.0.0', 4574, loop=loop)
    server = loop.run_until_complete(coro)

    # Serve requests until CTRL+c is pressed
    print('Serving on {}'.format(server.sockets[0].getsockname()))
    try:
        loop.run_forever()
    except KeyboardInterrupt:
        pass

    # Close the server
    server.close()
    loop.run_until_complete(server.wait_closed())
    loop.close()


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#  -*- coding: utf-8 -*-
import os
import sys
from setuptools import setup
from setuptools import find_packages

version = '1.4'

install_requires = []
test_requires = [
    'pytest',
    'pytest-asyncio<0.6.0' if sys.version_info < (3, 5) else 'pytest-asyncio',
    'coverage',
    'coveralls',
]


def read(*rnames):
    return open(os.path.join(os.path.dirname(__file__), *rnames)).read()


setup(
    name='panoramisk',
    version=version,
    description="asyncio based library to play with asterisk",
    long_description=read('README.rst'),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Communications :: Telephony',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    keywords=['asyncio', 'asterisk', 'voip'],
    author='Gael Pasgrimaud',
    author_email='gael@gawel.org',
    url='https://github.com/gawel/panoramisk/',
    license='MIT license',
    packages=find_packages(exclude=['docs', 'tests']),
    include_package_data=True,
    zip_safe=False,
    install_requires=install_requires,
    tests_require=test_requires,
    extras_require={
        'test': test_requires,
    },
    entry_points='''
    [console_scripts]
    panoramisk = panoramisk.command:main
    '''
)



from pprint import pprint
import asyncio
from panoramisk import Manager
import sys


async def ping(lp, username, secret):
    manager = Manager(loop=lp,
                      host='127.0.0.1', port=5038,
                      username=username, secret=secret,
                      forgetable_actions=('login',))
    await manager.connect()
    while True:
        p = await manager.send_action({'Action': 'ping'})
        # p = await manager.send_action({'Action': 'SIPpeers'})
        pprint(p)
        await asyncio.sleep(1)
    manager.close()


def main():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(ping(loop, sys.argv[1], sys.argv[2]))
    loop.close()


if __name__ == '__main__':
    main()



from __future__ import unicode_literals
import asyncio
from unittest import mock

from . import manager
from . import utils


MagicMock = mock.MagicMock
patch = mock.patch
call = mock.call


class AMIProtocol(manager.AMIProtocol):

    debug_count = [0]

    def connection_made(self, transport):
        super(AMIProtocol, self).connection_made(transport)
        self.transport = MagicMock()

    def send(self, data, as_list=False):
        utils.IdGenerator.reset(uid='transaction_uid')
        future = super(AMIProtocol, self).send(data, as_list=as_list)
        if getattr(self.factory, 'stream', None) is not None:
            with open(self.factory.stream, 'rb') as fd:
                for resp in fd.read().split(b'\n\n'):
                    self.data_received(resp + b'\n\n')
                    if future.done():
                        break
            if not future.done():  # pragma: no cover
                print(self.responses)
                raise AssertionError("Future's result was never set")
        return future


class Manager(manager.Manager):

    fixtures_dir = None
    defaults = manager.Manager.defaults.copy()

    def __init__(self, **config):
        self.defaults.update(
            protocol_factory=AMIProtocol,
            stream=None)
        super(Manager, self).__init__(**config)

        self.stream = self.config.get('stream')

        if self.loop is None:
            self.loop = asyncio.get_event_loop()

        protocol = AMIProtocol()
        protocol.factory = manager
        protocol.connection_made(mock.MagicMock())
        future = asyncio.Future()
        future.set_result((mock.MagicMock(), protocol))
        self.protocol = protocol
        self.connection_made(future)

        utils.IdGenerator.reset(uid='transaction_uid')
        utils.EOL = '\n'



import asyncio
import logging
import os

from panoramisk import Manager, Message


manager = Manager(
    host=os.getenv('AMI_HOST', '127.0.0.1'),
    port=os.getenv('AMI_PORT', 5038),
    username=os.getenv('AMI_USERNAME', 'user'),
    secret=os.getenv('AMI_SECRET', 'password'),
    ping_delay=10,  # Delay after start
    ping_interval=10,  # Periodically ping AMI (dead or alive)
    reconnect_timeout=2,  # Timeout reconnect if connection lost
)


def on_connect(mngr: Manager):
    logging.info(
        'Connected to %s:%s AMI socket successfully' %
        (mngr.config['host'], mngr.config['port'])
    )


def on_login(mngr: Manager):
    logging.info(
        'Connected user:%s to AMI %s:%s successfully' %
        (mngr.config['username'], mngr.config['host'], mngr.config['port'])
    )


def on_disconnect(mngr: Manager, exc: Exception):
    logging.info(
        'Disconnect user:%s from AMI %s:%s' %
        (mngr.config['username'], mngr.config['host'], mngr.config['port'])
    )
    logging.debug(str(exc))


async def on_startup(mngr: Manager):
    await asyncio.sleep(0.1)
    logging.info('Something action...')


async def on_shutdown(mngr: Manager):
    await asyncio.sleep(0.1)
    logging.info(
        'Shutdown AMI connection on %s:%s' % (mngr.config['host'], mngr.config['port'])
    )


@manager.register_event('*')  # Register all events
async def ami_callback(mngr: Manager, msg: Message):
    if msg.Event == 'FullyBooted':
        print(msg)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    manager.on_connect = on_connect
    manager.on_login = on_login
    manager.on_disconnect = on_disconnect
    manager.connect(run_forever=True, on_startup=on_startup, on_shutdown=on_shutdown)



from panoramisk import testing
import asyncio
import pytest


@pytest.fixture
def conn(request, event_loop):

    def callback(*args):
        pass

    loop = asyncio.get_event_loop_policy().new_event_loop()
    manager = testing.Manager(loop=loop)
    manager.register_event('Peer*', callback)
    return manager.protocol


def test_received(conn):
    conn.data_received(b'Event: None\r\n\r\n')
    conn.data_received(b'Event: PeerStatus\r\nPeer: gawel\r\n\r\n')
    conn.data_received(b'Response: Follows\r\nPeer: gawel\r\n\r\n')
    conn.data_received(b'Response: Success\r\nPing: Pong\r\n\r\n')


def test_send(conn):
    assert isinstance(conn.send({}), asyncio.Future)
